//
//  EAGLView.m
//  C3DConverterTest
//
//  Created by Rob Bajorek on 5/9/09.
//  Copyright SakariTech Inc. 2009. All rights reserved.
//

#import <QuartzCore/QuartzCore.h>
#import <OpenGLES/EAGLDrawable.h>

#import "EAGLView.h"
#import "4 shapes_iphone.h"

#define USE_DEPTH_BUFFER 1

// A class extension to declare private methods
@interface EAGLView ()

@property (nonatomic, retain) EAGLContext *context;
@property (nonatomic, assign) NSTimer *animationTimer;

- (BOOL) createFramebuffer;
- (void) destroyFramebuffer;

@end


@implementation EAGLView

@synthesize context;
@synthesize animationTimer;
@synthesize animationInterval;


// You must implement this method
+ (Class)layerClass {
  return [CAEAGLLayer class];
}


//The GL view is stored in the nib file. When it's unarchived it's sent -initWithCoder:
- (id)initWithCoder:(NSCoder*)coder {
  
  if ((self = [super initWithCoder:coder])) {
    // Get the layer
    CAEAGLLayer *eaglLayer = (CAEAGLLayer *)self.layer;
    
    eaglLayer.opaque = YES;
    eaglLayer.drawableProperties = [NSDictionary dictionaryWithObjectsAndKeys:
                                    [NSNumber numberWithBool:NO], kEAGLDrawablePropertyRetainedBacking, kEAGLColorFormatRGBA8, kEAGLDrawablePropertyColorFormat, nil];
    
    context = [[EAGLContext alloc] initWithAPI:kEAGLRenderingAPIOpenGLES1];
    
    if (!context || ![EAGLContext setCurrentContext:context]) {
      [self release];
      return nil;
    }
    
    // This is our max framerate
    animationInterval = 1.0 / 60.0;
  }
  return self;
}

// Use this method to set up multiple textures.
- (void) setupTexture: (NSString *) texture ID: (GLuint *) textureID {
  CGImageRef spriteImage;
  CGContextRef spriteContext;
  GLubyte *spriteData;
	size_t	width, height;
  
  // * Creates a Core Graphics image from an image file
	spriteImage = [UIImage imageNamed: texture].CGImage;
	// Get the width and height of the image
	width = CGImageGetWidth(spriteImage);
	height = CGImageGetHeight(spriteImage);
	// Texture dimensions must be a power of 2. If you write an application that allows users to supply an image,
	// you'll want to add code that checks the dimensions and takes appropriate action if they are not a power of 2.
  
	if(spriteImage) {
		// Allocated memory needed for the bitmap context
		spriteData = (GLubyte *) malloc(width * height * 4);
		// Uses the bitmatp creation function provided by the Core Graphics framework. 
		spriteContext = CGBitmapContextCreate(spriteData, width, height, 8, width * 4, CGImageGetColorSpace(spriteImage), kCGImageAlphaPremultipliedLast);
		// After you create the context, you can draw the sprite image to the context.
		CGContextDrawImage(spriteContext, CGRectMake(0.0, 0.0, (CGFloat)width, (CGFloat)height), spriteImage);
		// You don't need the context at this point, so you need to release it to avoid memory leaks.
		CGContextRelease(spriteContext);
		
		// Use OpenGL ES to generate a name for the texture.
		glGenTextures(1, textureID);
		// Bind the texture name. 
		glBindTexture(GL_TEXTURE_2D, *textureID);
		// Specify a 2D texture image, providing a pointer to the image data in memory
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, spriteData);
		// Release the image data
		free(spriteData);
		
		// Set the texture parameters to use a minifying  filter and a linear filer (weighted average)
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		
		// Enable use of the texture
		glEnable(GL_TEXTURE_2D);
    // Set a blending function to use
		glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
    // Enable blending
		glEnable(GL_BLEND);
    NSLog([NSString stringWithFormat: @"glError code: (%i)", glGetError()]);
    
	}
}

- (void) setupObjectTexture {
  [self setupTexture: @"pure_colors.jpg" ID: &objectTextureID];
}

// Performs one-time setup for OpenGL in our app.  Depnding on how complicated
// your app is, you may need to move some of these into your draw loop.
- (void) setupView {
  // Bind our framebuffer and set the view port to the iPhone screen size.
  [EAGLContext setCurrentContext:context];
  glBindFramebufferOES(GL_FRAMEBUFFER_OES, viewFramebuffer);
  glViewport(0, 0, backingWidth, backingHeight);
  
  // Set up the viewing volume and orthographic mode.
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrthof(-2.0f, 2.0f, -2.5f, 2.5f, -2.0f, 2.0f);
  
  // Clear the modelview matrix
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  
  // Enable the states our header file objects use.
  glEnableClientState(GL_VERTEX_ARRAY);
  glEnableClientState(GL_TEXTURE_COORD_ARRAY);
  glEnableClientState(GL_NORMAL_ARRAY);

  // Enable depth testing. Without this your back faces will may show through
  // the front ones.
  glEnable(GL_DEPTH_TEST);

  /* Lighting */
  
  // Light features
  GLfloat light_ambient[]= { 1.2f, 1.2f, 1.2f, 1.0f };
  GLfloat light_diffuse[]= { 1.0f, 1.0f, 1.0f, 0.0f };
  GLfloat light_specular[]= { 1.0f, 1.0f, 1.0f, 0.0f };
  // Set up light 0
  glLightfv (GL_LIGHT0, GL_AMBIENT, light_ambient);
  glLightfv (GL_LIGHT0, GL_DIFFUSE, light_diffuse);
  glLightfv (GL_LIGHT0, GL_SPECULAR, light_specular);
  
  // Material features
  // Note that "GL_FRONT" is not valid for arg0 on the iPhone (as of 2.2.1).
  GLfloat mat_specular[] = { 0.5, 0.5, 0.5, 1.0 };
  GLfloat mat_shininess[] = { 120.0 };
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, mat_specular);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, mat_shininess);
  
  // Set up shading model.  This should match your models' shading or your normals
  // will be wrong.
  glShadeModel (GL_SMOOTH);
  
  //  NSLog([NSString stringWithFormat: @"glError code: (%i)", glGetError()]);
  
  // Enable lighting and lights
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  
  // Place the light
  GLfloat light0_position[] = { 2.0, 5.0, 5.0, 1.0 };
  glLightfv(GL_LIGHT0, GL_POSITION, light0_position);
}

- (void)drawView {
  // Set matrix mode to modelview
  glMatrixMode(GL_MODELVIEW);
  
  // Rotate the view. Since we don't load identity here, the rotations will
  // stack each frame.
  glRotatef(0.75f, 1.0f , 0.0f, 0.0f);
  glRotatef(2.0f, 0.0f, 1.0f, 0.0f);

  // Clear the buffers with your favorite color. That color will be the background.
  // You must clear the depth buffer as well or there will be a mess.
  glClearColor(0.1f, 0.1f, 0.1f, 1.0f);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  
  // Draw the cube
  glVertexPointer(3, GL_FLOAT, 0, Box_vertex_coordinates);
  glTexCoordPointer(2, GL_FLOAT, 0, Box_texture_coordinates);
  glNormalPointer(GL_FLOAT, 0, Box_normal_vectors);
  glDrawArrays(GL_TRIANGLES, 0, Box_vertex_array_size);

  // Draw the cylinder
  glVertexPointer(3, GL_FLOAT, 0, Cylinder_vertex_coordinates);
  glTexCoordPointer(2, GL_FLOAT, 0, Cylinder_texture_coordinates);
  glNormalPointer(GL_FLOAT, 0, Cylinder_normal_vectors);
  glDrawArrays(GL_TRIANGLES, 0, Cylinder_vertex_array_size);

  // Draw the cone
  glVertexPointer(3, GL_FLOAT, 0, Cone_vertex_coordinates);
  glTexCoordPointer(2, GL_FLOAT, 0, Cone_texture_coordinates);
  glNormalPointer(GL_FLOAT, 0, Cone_normal_vectors);
  glDrawArrays(GL_TRIANGLES, 0, Cone_vertex_array_size);

  // Draw the sphere
  glVertexPointer(3, GL_FLOAT, 0, Ball_vertex_coordinates);
  glTexCoordPointer(2, GL_FLOAT, 0, Ball_texture_coordinates);
  glNormalPointer(GL_FLOAT, 0, Ball_normal_vectors);
  glDrawArrays(GL_TRIANGLES, 0, Ball_vertex_array_size);
  
  // Done, render to screen
  glBindRenderbufferOES(GL_RENDERBUFFER_OES, viewRenderbuffer);
  [context presentRenderbuffer:GL_RENDERBUFFER_OES];
}

// The setup for our GL view is here.
- (void)layoutSubviews {
  [EAGLContext setCurrentContext:context];
  [self destroyFramebuffer];
  [self createFramebuffer];
  [self setupObjectTexture];
  [self setupView];
  [self drawView];
}


- (BOOL)createFramebuffer {
  
  glGenFramebuffersOES(1, &viewFramebuffer);
  glGenRenderbuffersOES(1, &viewRenderbuffer);
  
  glBindFramebufferOES(GL_FRAMEBUFFER_OES, viewFramebuffer);
  glBindRenderbufferOES(GL_RENDERBUFFER_OES, viewRenderbuffer);
  [context renderbufferStorage:GL_RENDERBUFFER_OES fromDrawable:(CAEAGLLayer*)self.layer];
  glFramebufferRenderbufferOES(GL_FRAMEBUFFER_OES, GL_COLOR_ATTACHMENT0_OES, GL_RENDERBUFFER_OES, viewRenderbuffer);
  
  glGetRenderbufferParameterivOES(GL_RENDERBUFFER_OES, GL_RENDERBUFFER_WIDTH_OES, &backingWidth);
  glGetRenderbufferParameterivOES(GL_RENDERBUFFER_OES, GL_RENDERBUFFER_HEIGHT_OES, &backingHeight);
  
  // We are definitely using the depth buffer
  if (USE_DEPTH_BUFFER) {
    glGenRenderbuffersOES(1, &depthRenderbuffer);
    glBindRenderbufferOES(GL_RENDERBUFFER_OES, depthRenderbuffer);
    glRenderbufferStorageOES(GL_RENDERBUFFER_OES, GL_DEPTH_COMPONENT16_OES, backingWidth, backingHeight);
    glFramebufferRenderbufferOES(GL_FRAMEBUFFER_OES, GL_DEPTH_ATTACHMENT_OES, GL_RENDERBUFFER_OES, depthRenderbuffer);
  }
  
  if(glCheckFramebufferStatusOES(GL_FRAMEBUFFER_OES) != GL_FRAMEBUFFER_COMPLETE_OES) {
    NSLog(@"failed to make complete framebuffer object %x", glCheckFramebufferStatusOES(GL_FRAMEBUFFER_OES));
    return NO;
  }
  
  return YES;
}


- (void)destroyFramebuffer {
  
  glDeleteFramebuffersOES(1, &viewFramebuffer);
  viewFramebuffer = 0;
  glDeleteRenderbuffersOES(1, &viewRenderbuffer);
  viewRenderbuffer = 0;
  
  if(depthRenderbuffer) {
    glDeleteRenderbuffersOES(1, &depthRenderbuffer);
    depthRenderbuffer = 0;
  }
}


- (void)startAnimation {
  self.animationTimer = [NSTimer scheduledTimerWithTimeInterval:animationInterval target:self selector:@selector(drawView) userInfo:nil repeats:YES];
}


- (void)stopAnimation {
  self.animationTimer = nil;
}


- (void)setAnimationTimer:(NSTimer *)newTimer {
  [animationTimer invalidate];
  animationTimer = newTimer;
}


- (void)setAnimationInterval:(NSTimeInterval)interval {
  
  animationInterval = interval;
  if (animationTimer) {
    [self stopAnimation];
    [self startAnimation];
  }
}


- (void)dealloc {
  
  [self stopAnimation];
  
  if ([EAGLContext currentContext] == context) {
    [EAGLContext setCurrentContext:nil];
  }
  
  [context release];  
  [super dealloc];
}

@end
